from typing import Callable

from cvg_core.proper_procedures import SendReceiveProcedures

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionState, ConnectionObject

def command_send_and_receive(
    connection: ConnectionObject, 
    command: bytes,
    id: bytes | None = None
) -> PacketObject:
    return SendReceiveProcedures(connection).send_and_receive(
        PacketObject(command, PacketType.COMMAND, id),
        PacketType.RESPONSE
    )
    

def command_receive_and_send(
    connection: ConnectionObject,
    packet: PacketObject,
    id: bytes | None = None
) -> PacketObject:
    return SendReceiveProcedures(connection).receive_and_send(
        packet,
        PacketType.COMMAND,
        id
    )


def command_receive_into_and_send(
    connection: ConnectionObject,
    type: PacketType | None = None, 
    id: bytes | None = None
):
    procedures = SendReceiveProcedures(connection)
    
    def wrapper(func: Callable[[PacketObject], PacketObject], *args: any):
        packet = procedures.receive(id)
        result = func(packet, *args)
        
        if type:
            assert packet.type == type
        
        procedures.send(result)
        
        return result
    
    return wrapper


from enum import Enum
from typing import Callable
from dataclasses import dataclass, field

from Crypto.Cipher import AES

from cryptography.hazmat.backends import default_backend

from cryptography.hazmat.primitives import serialization, hashes

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.kdf.hkdf import HKDF


@dataclass
class ECDHObject:
    public_key: None | bytes | ec.EllipticCurvePublicKey = field(
        default=None
    )
    
    private_key: None | bytes | ec.EllipticCurvePrivateKey = field(
        default=None
    )
    
    curve: ec.EllipticCurve = field(default_factory=ec.SECP521R1)
    backend: Callable = field(default_factory=default_backend)
    
    hkdf_length: int = field(default=32)
    hkdf_salt: bytes = field(default=b"")
    hkdf_info: bytes = field(default=b"")
    
    def __post_init__(self):
        _public_key_type = type(self.public_key)
        _private_key_type = type(self.private_key)

        if _public_key_type is bytes:
            self.public_key = serialization.load_pem_public_key(
                self.public_key
            )
        
        if self.public_key is None:
            if _private_key_type is bytes:
                self.private_key = serialization.load_pem_private_key(
                    self.private_key
                )
            elif self.private_key is None:
                self.private_key = ec.generate_private_key(
                    self.curve, self.backend
                )

                self.public_key = self.private_key.public_key()

    def to_public_pem(self) -> bytes:
        return self.public_key.public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo
        )
    
    def derive_secret(self, public_key: bytes | ec.EllipticCurve) -> bytes:
        if type(public_key) is bytes:
            public_key = serialization.load_pem_public_key(public_key)
        
        return self.private_key.exchange(
            ec.ECDH(), 
            public_key
        )
    
    def derive_aes_key(self, public_key: bytes | ec.EllipticCurve) -> bytes:    
        return HKDF(
            algorithm=hashes.SHA256(),
            length=self.hkdf_length,
            salt=self.hkdf_salt,
            info=self.hkdf_info,
            backend=self.backend
        ).derive(self.derive_secret(public_key))


from cvg_core.proper_procedures import SendReceiveProcedures

from cvg_core.procedures.establish_connection import establish_connection

from cvg_core.objects.crypto_object.ecdh_object import ECDHObject

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionObject




import sys

from time import sleep
from threading import Thread

from socket import socket, AF_INET, SOCK_STREAM

from cvg_core import SendReceiveProcedures, PacketType, ConnectionType, PacketObject, ConnectionObject, establish_connection

password = b"bytes"

def client_example():
    client_connection = ConnectionObject(
        address=("127.0.0.1", 5000),
        socket=socket(AF_INET, SOCK_STREAM),
        type=ConnectionType.CLIENT_TO_SERVER,
        encryption_enabled=True
    )

    client_connection.socket.connect(
        client_connection.address
    )

    establish_connection(client_connection, password)

    client_procedures = SendReceiveProcedures(client_connection)

    # At this point it's up to on what you want to do.
    command_result_a = client_procedures.send_and_receive(
        PacketObject(b"hello", PacketType.COMMAND, id=b"\x03"),
        PacketType.RESPONSE
    )

    command_result_b = client_procedures.send_and_receive(
        PacketObject(b"hello", PacketType.COMMAND, id=b"\x03"),
        PacketType.RESPONSE
    )

    sleep(0.1) # sometimes the prints stack on each other.

    print("[client] command result a:", command_result_a)
    print("[client] command result b:", command_result_b)


# [Server Implementation]
def server_example():
    server_socket = socket(AF_INET, SOCK_STREAM)
    server_socket.bind(("127.0.0.1", 5000))
    server_socket.listen(1)

    client_connection = ConnectionObject(
        *server_socket.accept(), 
        type=ConnectionType.SERVER_TO_CLIENT,
        encryption_enabled=True
    ) 
        
    establish_connection(client_connection, password)

    client_procedures = SendReceiveProcedures(client_connection)

    # At this point it's up to on what you want to do.

    # Function wrapping on receive.
    @client_procedures.receive_into_and_send(PacketType.COMMAND)
    def command(packet: PacketObject):
        print("[server] client command received", packet)
        
        if packet.payload.startswith(b"hello"):
            return PacketObject(b"world_a", PacketType.RESPONSE, packet.id)

    # Or send a premade packet upon receive.
    command_packet = client_procedures.receive_and_send(
        PacketObject(b"world_b", PacketType.RESPONSE),
        PacketType.COMMAND
    )
    
    print("[server] client command received", command_packet)

Thread(target=server_example).start()
sleep(1)
client_example()


import hashlib

from typing import Callable

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionState, ConnectionObject

ERR_MSG_ID_MISMATCH = "PacketObject.id mismatch! Expected {0} got {0}"
ERR_MSG_STREAM_NOT_READY = "Connection is not ready for stream!"

def stream_receive(
    connection: ConnectionObject, size: int,
    id: bytes | None = b"\xff"
) -> PacketObject:
    compiled_payload = b""
    stream_packet = PacketObject(b"", PacketType.STREAM_DATA, id)
    
    packet = send_and_receive(
        connection, 
        stream_packet,
        PacketType.STREAM_DATA
    )
    
    # TODO packet can return None handle interrupt
    
    compiled_payload += packet.payload
    
    while True:
        packet = send_and_receive(connection, stream_packet)
        
        # TODO packet can return None handle interrupt
        
        if packet.type is PacketType.STREAM_DATA:
            compiled_payload += packet.payload
        elif packet.type is PacketType.STREAM_END:
            break

    # TODO add err msg
    assert len(compiled_payload) == size

    md5_sum = hashlib.md5()
    md5_sum.update(compiled_payload)
    
    received_checksum = packet.payload
    checksum = md5_sum.digest()
    
    send(
        connection,
        PacketObject(checksum, PacketType.STREAM_END)
    )
    
    assert checksum == received_checksum, "Receive stream checksum failed!"
    
    return PacketObject(compiled_payload)


def stream_send(connection: ConnectionObject, packet: PacketObject):
    raw_packet = packet.to_bytes()
    packet_size = packet.get_size()
    
    chunk_count = int(packet_size / 4094)
    chunk_remainder = packet_size % 4094
    
    chunk_list = [
        raw_packet[index : index + 4094] for index in range(
            0, 
            chunk_count * 4094,
            4094
        ) 
    ]
    
    chunk_list.append(raw_packet[packet_size - chunk_remainder: packet_size])
    
    ready_packet = send_and_receive(
        connection,
        PacketObject(
            packet_size.to_bytes(8, "big"), 
            PacketType.STREAM_START, 
            packet.id
        ),
        PacketType.STREAM_DATA
    )
    
    # TODO packet can return None handle interrupt
    
    for chunk in chunk_list:
        status_packet = send_and_receive(
            connection,
            PacketObject(chunk, PacketType.STREAM_DATA, packet.id)
        )
        
        # TODO stream interrupt, status_packet can return None!
        # TODO check if status is not stream_data, could be cancelled/failed.
        
    # TODO is this really the cleanest I can do?
    md5_sum = hashlib.md5()
    md5_sum.update(raw_packet)
    
    checksum = md5_sum.digest()

    received_checksum = send_and_receive(
        connection,
        PacketObject(
            checksum, 
            PacketType.STREAM_END, 
            packet.id
        )
    ).payload
    
    # TODO move err msg
    assert checksum == received_checksum, "Send stream checksum failed!"
    

def receive(
    connection: ConnectionObject, 
    receive_type: PacketType | None = None,
    receive_id: bytes | None = None
) -> PacketObject | None:
    packet: PacketObject | None = None
    
    try:
        packet = PacketObject(connection.socket.recv(4096))
    except Exception as _:
        pass
    
    if packet and receive_id is not None:
        assert packet.id == receive_id, ERR_MSG_ID_MISMATCH.format(
            receive_id, 
            packet.id
        )

    # TODO move err msg
    if packet and receive_type:
        assert packet.type is receive_type, f"expected {receive_type} got {packet.type}"
    
    if packet is not None and packet.type is PacketType.STREAM_START:
        packet = stream_receive(
            connection,
            int.from_bytes(packet.payload, "big"),
            packet.id
        )
    
    return packet


def send(connection: ConnectionObject, packet: PacketObject):
    if packet.get_size() > 4096:
        stream_send(connection, packet)
    else:
        connection.socket.send(packet.to_bytes())


def send_and_receive(
    connection: ConnectionObject, packet: PacketObject,
    receive_type: PacketType | None = None
) -> PacketObject | None:
    send(connection, packet)
    
    result = receive(connection, receive_type, packet.id)
    
    return result


def receive_and_send(
    connection: ConnectionObject,
    send_packet: PacketObject,
    receive_type: PacketType | None = None,
    receive_id: bytes | None = None
) -> PacketObject | None:
    result = receive(connection, receive_type, receive_id)

    send_packet.id = result.id
    
    send(connection, send_packet)
    
    return result


def receive_into_and_send(
    connection: ConnectionObject, 
    receive_type: PacketType | None = None, 
    receive_id: bytes | None = None
):
    def wrapper(func: Callable[[PacketObject], PacketObject], *args: any):
        packet = receive(connection, receive_type, receive_id)
        result = func(packet, *args)
        
        send(connection, result)
        
        return result
    
    return wrapper


from enum import Enum
from dataclasses import dataclass, field


class PacketType(Enum):
    UNKNOWN = b"\xff"
    
    GATEWAY = b"\x00"
    EXCHANGE = b"\x01"
    
    CRYPTO = b"\xe0"
    
    PASSWORD = b"\x02"
    
    GRANTED = b"\xa0"
    DENIED = b"\xa1"
    
    COMMAND = b"\xc0"
    RESPONSE = b"\xc1"
    
    STREAM_START = b"\xb0"
    STREAM_DATA = b"\xb1"
    STREAM_END = b"\xb2"


@dataclass
class PacketObject:
    payload: bytes = field(default=b"")
    type: PacketType = field(default=PacketType.UNKNOWN)
    
    id: bytes = field(default=b"\x00")
    size: int = field(default=2)
    
    def __post_init__(self):
        if self.type is PacketType.UNKNOWN:
            # TODO replace ifs with asserts
            if len(self.payload) < 2:
                raise Exception("Invalid payload length!")
            
            self.id = self.payload[0:1]
            
            try:
                self.type = PacketType(self.payload[1:2])
            except:
                raise Exception("Invalid packet type!")
            
            self.payload = self.payload[2::]
        elif self.id is None:
            self.id = b"\x00"

        self.size = self.get_size()

    def to_bytes(self) -> bytes:
        return self.id + self.type.value + self.payload
    
    def get_size(self) -> int:    
        return len(self.to_bytes())
    
    def get_payload_size(self) -> int:
        return self.get_size() - 2

    def set_payload(self, data: bytes):
        self.payload = data
        self.to_bytes()
    
    def add_payload(self, data: bytes):
        self.payload += data
        self.to_bytes()
        
    def clear_payload(self):
        self.payload = b""
        self.size = 2
        
    


import hashlib

from typing import Callable

from Crypto.Cipher import AES

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionState, ConnectionObject

from cvg_core.objects.crypto_object.ecdh_object import ECDHObject

from cvg_core.procedures.send_and_receive import send_and_receive, send, receive


def __derive_aes_key(connection: ConnectionObject):
    match connection.type:
        case ConnectionType.SERVER_TO_CLIENT:
            return connection.server_crypto.derive_aes_key(
                connection.client_crypto.public_key
            )
        case ConnectionType.CLIENT_TO_SERVER:
            return connection.client_crypto.derive_aes_key(
                connection.server_crypto.public_key
            )


def __to_public_pem(connection: ConnectionObject):
    match connection.type:
        case ConnectionType.SERVER_TO_CLIENT:
            return connection.server_crypto.to_public_pem()
        case ConnectionType.CLIENT_TO_SERVER:
            return connection.client_crypto.to_public_pem()


def decrypt_packet(
    connection: ConnectionObject, packet: PacketObject
) -> PacketObject:
    return PacketObject(
        AES.new(
            key=__derive_aes_key(connection),
            mode=AES.MODE_EAX,
            nonce=packet.payload[0:16]
        ).decrypt(packet.payload[16::])
    )


def encrypt_packet(
    connection: ConnectionObject, packet: PacketObject
) -> PacketObject:
    cipher = AES.new(
        key=__derive_aes_key(connection),
        mode=AES.MODE_EAX
    )
    
    return PacketObject(
        cipher.nonce + cipher.encrypt(packet.to_bytes()),
        PacketType.CRYPTO
    )


def crypto_exchange(connection: ConnectionObject):
    if connection.type is ConnectionType.SERVER_TO_CLIENT:
        if connection.encryption_enabled:
            client_pem = send_and_receive(
                connection,
                PacketObject(
                    __to_public_pem(connection),
                    PacketType.EXCHANGE
                ),
                PacketType.EXCHANGE
            )
        
            connection.client_crypto = ECDHObject(client_pem.payload)
            
            # TODO confirm crypto exchange. we just assume it's a success rn.
            crypto_send(
                connection, 
                PacketObject(b"Hello", PacketType.EXCHANGE, client_pem.id)
            )
        else:
            send(
                connection,
                PacketObject(b"Hello", PacketType.EXCHANGE)
            )
            
            connection.encryption_enabled = False
        
    else:
        server_exchange = receive(connection, PacketType.EXCHANGE)
        
        if server_exchange.payload.startswith(b"-----BEGIN PUBLIC KEY-----"):
            connection.client_crypto = ECDHObject()
            connection.server_crypto = ECDHObject(server_exchange.payload)
            
            send(
                connection,
                PacketObject(
                    __to_public_pem(connection), 
                    PacketType.EXCHANGE,
                    server_exchange.id
                )
            )
            
            # TODO confirm crypto exchange. we just assume it's a success rn.
            crypto_receive(connection, PacketType.EXCHANGE, server_exchange.id)
            
            connection.encryption_enabled = True
        elif server_exchange.payload == b"Hello":
            connection.encryption_enabled = False
        

def crypto_send(connection: ConnectionObject, packet: PacketObject):
    send(connection, encrypt_packet(connection, packet))
    

def crypto_receive(
    connection: ConnectionObject,
    receive_type: PacketType | None = None,
    id: None | bytes = None
) -> PacketObject | None:
    packet = receive(connection, PacketType.CRYPTO)
    
    result = decrypt_packet(connection, packet)
    
    if id:
        assert result.id == id
    
    if receive_type:
        # TODO move err msg, do not remove this. This assert is required.
        assert result.type is receive_type, f"got {result.type} expected {receive_type}"
    
    return result
        

def crypto_send_and_receive(
    connection: ConnectionObject,
    packet: PacketObject,
    receive_type: PacketType | None = None
) -> PacketObject | None:
    crypto_send(connection, packet)
    return crypto_receive(connection, receive_type, packet.id)


def crypto_receive_and_send(
    connection: ConnectionObject,
    send_packet: PacketObject,
    receive_type: PacketType | None = None,
    receive_id: bytes | None = None,
) -> PacketObject | None:
    result = crypto_receive(connection, receive_type, receive_id)

    send_packet.id = result.id
    
    crypto_send(connection, send_packet)
    
    return result


def crypto_receive_into_and_send(
    connection: ConnectionObject, 
    type: PacketType | None = None, 
    id: bytes | None = None
):
    def wrapper(func: Callable[[PacketObject], PacketObject], *args: any):
        packet = crypto_receive(connection, type, id)
        result = func(packet, *args)
        
        crypto_send(connection, result)
        
        return result
    
    return wrapper


from typing import Callable

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionState, ConnectionObject

from cvg_core.proper_procedures import SendReceiveProcedures
from cvg_core.procedures.crypto_send_and_receive import crypto_exchange


def __fab_granted(
    packet: PacketObject | bytes, 
    payload: bytes | None = b""
) -> PacketObject:
    if type(packet) is bytes:  
        return PacketObject(payload, PacketType.GRANTED, packet)
    else:
        return PacketObject(payload, PacketType.GRANTED, packet.id)


def __fab_denied(
    packet: PacketObject | bytes, 
    payload: bytes | None = b""
) -> PacketObject:
    if type(packet) is bytes:  
        return PacketObject(payload, PacketType.DENIED, packet)
    else:
        return PacketObject(payload, PacketType.DENIED, packet.id)


def __fab_password(
    packet: bytes | PacketObject,
    password: bytes | None = b""
):
    if type(packet) is bytes:  
        return PacketObject(password, PacketType.PASSWORD, packet)
    else:
        return PacketObject(password, PacketType.PASSWORD, packet.id)


def __exchange_password(
    packet: PacketObject,
    compare_password: bytes, procedures: SendReceiveProcedures
) -> PacketObject:
    received_password: PacketObject = procedures.send_and_receive(
        __fab_password(packet)
    )

    if received_password.payload == compare_password:
        procedures.connection.established = True
        return __fab_granted(received_password)
    else:
        return __fab_denied(received_password)


def establish_connection(
    connection: ConnectionObject, 
    password: bytes | None = None
):  
    #f connection.encryption_enabled:
    crypto_exchange(connection)
    
    procedures = SendReceiveProcedures(connection)
      
    if connection.type is ConnectionType.SERVER_TO_CLIENT:
        gateway_packet: PacketObject
        
        if password:
            gateway_packet = procedures.receive_into_and_send(
                PacketType.GATEWAY
            )(
                __exchange_password,
                password,
                procedures
            )
        else:
            gateway_packet = procedures.receive_into_and_send(
                PacketType.GATEWAY
            )(__fab_granted)
        
        assert gateway_packet.type is PacketType.GRANTED
        
    elif connection.type is ConnectionType.CLIENT_TO_SERVER:
        entry_response: PacketObject = procedures.send_and_receive(
            PacketObject(b"", PacketType.GATEWAY)
        )
        
        if entry_response.type is PacketType.PASSWORD:            
            password_response: PacketObject = procedures.send_and_receive(
                __fab_password(entry_response, password)
            )
            
            if password_response.type is PacketType.GRANTED:
                connection.established = True
            else:
                raise Exception("Incorrect password!")
            
        elif entry_response.type is PacketType.GRANTED:
            connection.established = True


from enum import Enum
from dataclasses import dataclass, field

from socket import socket as _socket


from cvg_core.objects.crypto_object.ecdh_object import ECDHObject


class ConnectionState(Enum):
    GREETING = "greeting"
    AUTHORIZED = "authorized"
    RESPONDING = "responding"
    STREAMING = "streaming"


class ConnectionType(Enum):
    CLIENT_TO_SERVER = "client_to_server"
    SERVER_TO_CLIENT = "server_to_client"
    UNKNOWN = "unknown"


@dataclass
class ConnectionObject:
    socket: _socket = field(default=None)
    address: tuple[str, int] = field(default=("", -1))
    
    type: ConnectionType = field(default=ConnectionType.UNKNOWN)
    
    encryption_enabled: bool = field(default=True)
    
    server_crypto: ECDHObject | None = field(default=None)
    client_crypto: ECDHObject | None = field(default=None)
    
    state: ConnectionState = field(default=ConnectionState.GREETING)
    established: bool = field(default=False)
    
    def __post_init__(self):
        if self.encryption_enabled:
            match self.type:
                case ConnectionType.CLIENT_TO_SERVER:
                    self.client_crypto = ECDHObject()
                case ConnectionType.SERVER_TO_CLIENT:
                    self.server_crypto = ECDHObject()
                case _:
                    pass


from typing import Any, Callable
from dataclasses import dataclass, field

from cvg_core.objects.network_object.packet_object import PacketType, PacketObject
from cvg_core.objects.network_object.connection_object import ConnectionType, ConnectionObject

from cvg_core.procedures.send_and_receive import send_and_receive, receive_and_send, send, receive, receive_into_and_send
from cvg_core.procedures.crypto_send_and_receive import crypto_send_and_receive, crypto_receive_and_send, crypto_send, crypto_receive, crypto_receive_into_and_send


@dataclass
class SendReceiveProcedures:
    connection: ConnectionObject
    
    __receive_into_and_send: Callable[
        [ConnectionObject, PacketType | None, bytes | None],
        Callable
    ] = field(default=receive_into_and_send)
    
    __send_and_receive: Callable[
        [ConnectionObject, PacketObject, PacketType | None], PacketObject | None
    ] = field(default=send_and_receive)
    
    __receive_and_send: Callable[
        [ConnectionObject, PacketObject, PacketType | None, bytes | None], 
        PacketObject | None
    ] = field(default=receive_and_send)
    
    __receive: Callable[
        [ConnectionObject, bytes | None], PacketObject | None
    ] = field(default=receive)
    
    __send: Callable[
        [ConnectionObject, PacketObject], PacketObject | None
    ] = field(default=send)

    def __post_init__(self):
        if self.connection and self.connection.encryption_enabled:
            self.__receive_into_and_send = crypto_receive_into_and_send
            
            self.__send_and_receive = crypto_send_and_receive
            self.__receive_and_send = crypto_receive_and_send
            
            self.__receive = crypto_receive
            self.__send = crypto_send
    
    def receive_into_and_send(
        self, 
        receive_type: PacketType | None = None, 
        receive_id: bytes | None = None
    ) -> PacketObject:
        return self.__receive_into_and_send(
            self.connection,
            receive_type,
            receive_id
        )
    
    def send_and_receive(
        self,
        packet: PacketObject,
        receive_type: PacketType | None = None
    ) -> PacketObject:
        return self.__send_and_receive(
            self.connection,
            packet,
            receive_type,
        )
        
    def receive_and_send(
        self,
        send_packet: PacketObject,
        receive_type: PacketType | None = None,
        receive_id: bytes | None = None
    ) -> PacketObject:
        return self.__receive_and_send(
            self.connection,
            send_packet,
            receive_type,
            receive_id
        )
        
    def receive(
        self,
        receive_type: PacketType | None = None,
        receive_id: bytes | None = None
    ) -> PacketObject:
        return self.__receive(
            self.connection,
            receive_type,
            receive_id
        )
        
    def send(
        self,
        packet: PacketObject
    ):
        self.__send(self.connection, packet)